unit AlphaZoom;

interface

uses
  Windows, Classes, Graphics, Controls, Forms {screen}
  , messages
  , TeEngine, TeeProcs, Chart
  ,TeCanvas//PointInRect
  , math
  ;


type
  //this is an extra bonus: make left and rignht mouse buttons interchangeable
  //(currently, Right mouse button stops working outside TChart ChartRect)
  //override TChart events to allow MouseUp event captured outside Chart->
  TChart = class(Chart.TChart)
  private
    procedure WMRButtonDown(var Message: TWMRButtonDown); message WM_RBUTTONDOWN;
    procedure WMRButtonUp(var Message: TWMRButtonUp); message WM_RBUTTONUP;
  protected
    FOnScrolled    : TNotifyEvent; //add onScrolled event
  public
    property OnScrolled :TNotifyEvent read FOnScrolled write FOnScrolled;
  end;

type
  TAlphaZoomTool=class(TTeeCustomTool)
  private
    PZoom : ^TTeeZoom;    //pointer to parent chart Zoom structure
    RSel  : TRect;        //zoom rectangle
    BmpBak: TBitmap;      //backup snapshot
    BmpNew: TBitmap;
    IZooming: Boolean;    //zooming
    IPanning: Boolean;    //zooming
    Panned  : Boolean;    //scrolled
    FBlendFactor: integer;//0..255
    FCursorZoom: TCursor; //use cursor while zooming
    FCursorPan : TCursor; //use cursor while panning
    FBackgrOnly: boolean; //activate zoom only if clicked background (skip all series)
    FSkipSeries: boolean; //activate zoom only if clicked outside of selected series (Use Series.Cursor<>crDefault to select series)


    procedure ActivateZoom;
    procedure DrawBlendedRect;
    procedure SetBlendFactor(Value: integer);
  protected
    procedure ChartMouseEvent( AEvent: TChartMouseEvent; AButton:TMouseButton; AShift: TShiftState; X, Y: Integer); override;
  public
    Constructor Create(AOwner:TComponent); override;
    Destructor Destroy; override;

    property BlendFactor: integer read FBlendFactor write SetBlendFactor;
    property CursorZoom: TCursor read FCursorZoom write FCursorZoom;
    property CursorPan : TCursor read FCursorPan  write FCursorPan;
    property BackgrOnly: boolean read FBackgrOnly write FBackgrOnly;
    property SkipSeries: boolean read FSkipSeries write FSkipSeries;
  end;

implementation



//allow right mouse release event outside of the chart bounds
procedure TChart.WMRButtonDown(var Message: TWMRButtonDown);
begin
     if csCaptureMouse in ControlStyle then MouseCapture := True;
     inherited;
end;
procedure TChart.WMRButtonUp(var Message: TWMRButtonUp);
begin
//     if csCaptureMouse in ControlStyle then MouseCapture := False;
     inherited;
     if csCaptureMouse in ControlStyle then MouseCapture := False;
end;


//======================================================================================================================
// TAlphaZoomTool->
//======================================================================================================================

Constructor TAlphaZoomTool.Create(AOwner: TComponent);
begin
  inherited;
  Self.ParentChart:=TChart(AOwner);

  PZoom:=@ParentChart.Zoom;

  //set default zoom colors & style->
  PZoom.Pen.Width:=1;
  PZoom.Pen.Color:=RGB(51, 153, 255);//clHighlight
  PZoom.Brush.Color:=RGB(51, 153, 255);//clHighlight;
  PZoom.Brush.Style:=bsSolid;//default

  FBlendFactor:=64;   //brush blend coefficient
  FBackgrOnly:=False; //standard zoom behaviour
  FSkipSeries:=False; //standard zoom behaviour
  FCursorZoom:=crDefault;
  FCursorPan :=crDefault;

  BmpBak:= TBitmap.Create;
  BmpNew:= TBitmap.Create;
end;

Destructor TAlphaZoomTool.Destroy;
begin
  BmpBak.Free;
  BmpNew.Free;
  inherited;
end;

procedure TAlphaZoomTool.ActivateZoom;
begin
     //make backup copy->
     BmpBak.Width:=ParentChart.Width;
     BmpBak.Height:=ParentChart.Height;
     BmpBak.PixelFormat:=pf32bit;
     BitBlt(BmpBak.Canvas.Handle, 0, 0, ParentChart.Width-1, ParentChart.Height-1, ParentChart.Canvas.Handle, 0, 0, srccopy); //restore chart canvas
end;

procedure TAlphaZoomTool.DrawBlendedRect; //draw blended selection region
var C, tmpClr: TColor;
    Bitmap: TBitmap;
    BlendFunction: TBlendFunction;
    DC: HDC;

  procedure FillLongword(var X; Count: Integer; Value: Longword);//from GR32_LowLevel.pas
  asm
  // EAX = X
  // EDX = Count
  // ECX = Value
       PUSH    EDI

       MOV     EDI,EAX  // Point EDI to destination
       MOV     EAX,ECX
       MOV     ECX,EDX
       TEST    ECX,ECX
       JS      @exit

       REP     STOSD    // Fill count dwords
  @exit:
       POP     EDI
  end;

begin

     BmpNew.Assign(BmpBak);//restore NewBmp

     //draw blended rect using Microsoft blending routine->
     if PZoom.Brush.Style <> bsClear then //not significant speedup
     with RSel do
     if (Right<>Left) and (Bottom<>Top) then begin
       Bitmap := TBitmap.Create;
       try
         Bitmap.PixelFormat := pf32Bit;
         Bitmap.SetSize(abs(Right-Left), abs(Bottom-Top) );

         C:=PZoom.Brush.Color;
         tmpClr:=$FF000000 + getRValue(C) shl 16 + getGValue(C) shl 8 + getBValue(C);//swap BGR->RGB

         FillLongword(Bitmap.ScanLine[Bitmap.Height - 1]^, Bitmap.Width * Bitmap.Height, tmpClr);

         BlendFunction.BlendOp := AC_SRC_OVER;
         BlendFunction.BlendFlags := 0;
         BlendFunction.SourceConstantAlpha := FBlendFactor;//0..255
         BlendFunction.AlphaFormat := AC_SRC_ALPHA;

         //Windows.
         AlphaBlend(BmpNew.Canvas.Handle, min(Left,Right), min(Top,Bottom), Bitmap.Width, Bitmap.Height,
                    Bitmap.Canvas.Handle, 0, 0, Bitmap.Width, Bitmap.Height, BlendFunction);
       finally
         Bitmap.Free;
       end;
     end;

     //draw frame on top->
     BmpNew.Canvas.Pen.Color:=PZoom.Pen.Color;
     BmpNew.Canvas.Pen.Width:=PZoom.Pen.Width;
     BmpNew.Canvas.Brush.Style:=bsClear;
     with BmpNew.Canvas, Rsel do Rectangle(Left,Top,Right,Bottom);

     //re-paint whole chart canvas->
     DC:=ParentChart.Canvas.Handle;
     BitBlt(DC, 0, 0, BmpNew.Width-1, BmpNew.Height-1, BmpNew.Canvas.Handle, 0, 0, srccopy);
end;

Procedure TAlphaZoomTool.ChartMouseEvent( AEvent: TChartMouseEvent; AButton:TMouseButton; AShift: TShiftState; X, Y: Integer);

   function BackgroundClicked(P: TPoint): boolean;
   var ClickedPart: TChartClickedPart;
   begin
     (ParentChart as TCustomChart).CalcClickedPart( Point(X,Y), ClickedPart);
     result := ClickedPart.Part <> cpChartRect;
   end;

   function MarkedSeriesClicked: boolean; //over series point with Series.Cursor<>crDefault
   begin
     result := ParentChart.Cursor <> crDefault;
   end;

begin

  Case AEvent of
    cmeDown: if AButton = PZoom.MouseButton then
             begin
               ParentChart.CancelMouse:=True; //kill original zoom

               if FBackgrOnly and BackgroundClicked(Point(X,Y)) then exit;
               if FSkipSeries and MarkedSeriesClicked then exit;

               if PZoom.Allow and (not IZooming) then begin
                 IZooming:=True;

                 if PZoom.Direction=tzdVertical   then X:=ParentChart.ChartRect.Left;
                 if PZoom.Direction=tzdHorizontal then Y:=ParentChart.ChartRect.Top;
                 RSel.Left:=X; RSel.Top:=Y; RSel.Right:=X; RSel.Bottom:=Y;//do not comment Right/Bottom part!
                 ActivateZoom;
                 Screen.Cursor:=FCursorZoom;//set zoom cursor style
               end;

             end//AButton=PZoom.MouseButton
             else

             if AButton = TChart(ParentChart).ScrollMouseButton then
             begin
               ParentChart.CancelMouse:=True; //kill native pan

               if ParentChart.AllowPanning = pmNone then exit;

               if not PointInRect(ParentChart.ChartRect, X,Y) then exit;
               if FBackgrOnly and BackgroundClicked(Point(X,Y)) then exit;
               if FSkipSeries and MarkedSeriesClicked then exit;
               if IZooming then exit;

               IPanning:=True;
               Panned:=False;
               ParentChart.CancelMouse:=False; //proceed to pan
               Screen.Cursor:=FCursorPan;//set zoom cursor style
             end;

    cmeMove: if IZooming then // zooming
             begin
               if PZoom.Direction=tzdVertical   then X:=ParentChart.ChartRect.Right;
               if PZoom.Direction=tzdHorizontal then Y:=ParentChart.ChartRect.Bottom;

               RSel.Right := X; RSel.Bottom := Y;
               DrawBlendedRect;
             end
             else

             if IPanning then // scrolling
             begin
               Panned:=True;
//               RSel.Right := X; RSel.Bottom := Y;
             end;

      cmeUp: if AButton=PZoom.MouseButton then
             begin
               RSel.Right := X; RSel.Bottom := Y;

               if IZooming then  // if Zooming
               begin
                 IZooming := False;
                 Screen.Cursor := crDefault;//restore Chart cursor

                 with RSel do
                 if (abs(Left - Right) < PZoom.MinimumPixels)
                 or (abs(Top - Bottom) < PZoom.MinimumPixels) then
                 begin
                   BitBlt(ParentChart.Canvas.Handle, 0, 0, BmpBak.Width-1, BmpBak.Height-1, BmpBak.Canvas.Handle, 0, 0, srccopy); //restore chart canvas
                   //ParentChart.Repaint;//slow on large data sets, do not use!
                   exit;
                 end;

                 with (ParentChart as TChart), RSel do
                 if ((Left < Right) and (Top < Bottom)) or PZoom.UpLeftZooms
                 then ZoomRect(RSel) //do zoom
                 else UndoZoom;
               end;

             end//AButton=PZoom.MouseButton
             else

             if AButton = TChart(ParentChart).ScrollMouseButton then begin

               if IPanning then IPanning:=False;

               if IZooming then begin
                 ParentChart.CancelMouse:=True; //kill original pan
                 exit;
               end;

               Screen.Cursor := crDefault;//restore Chart cursor
               if Assigned(TChart(ParentChart).OnScrolled) and Panned then TChart(ParentChart).FOnScrolled(self);

             end;

  end;


end;

procedure TAlphaZoomTool.SetBlendFactor(Value: integer);
begin
     if Value < 0   then  Value:=0 else
     if Value > 255 then  Value:=255;
     FBlendFactor:=Value;
end;


//======================================================================================================================
//<- TAlphaZoomTool
//======================================================================================================================

end.



//from TeeChart->
//  if not CheckMouseSeries(x,y) then { is mouse not over Series ? }
//     Cursor:=OriginalCursor; { change back original cursor }
//  if CancelMouse or Zoom.Active or IPanning.Active then Exit;

{
//type
//  TCustomTeePanelExtended=class(TeeProcs.TCustomTeePanelExtended) //not working...
//  public
//    Procedure DrawZoomRectangle;// override;
//  end;

procedure TCustomTeePanelExtended.DrawZoomRectangle;
var tmp:TColor;
Begin

  tmp:=ColorToRGB(GetBackColor);

  if Assigned(Zoom.Brush) then
     SetBrushCanvas((clWhite-tmp) xor Zoom.Brush.Color,Zoom.Brush,tmp)
  else
     InternalCanvas.Brush.Style:=bsClear;

  Zoom.Pen.Mode:=pmNotXor;

  With InternalCanvas do
  Begin
    AssignVisiblePenColor(Zoom.Pen,(clWhite-tmp) xor Zoom.Pen.Color);

    if (not Assigned(Zoom.Brush)) or
       (Zoom.Brush.Style=bsClear) then BackMode:=cbmTransparent;

    with Zoom do Rectangle(X0,Y0,X1,Y1);
  end;

end;
}

